@@ -4,7 +4,7 @@ import numpy as np
 import os
 from numba import njit, prange
 
-from .maxcut_tfim_util import get_cut, maxcut_hamming_cdf, opencl_context, sample_mag, bit_pick, init_bit_pick
+from .maxcut_tfim_util import get_cut, get_cut_base, maxcut_hamming_cdf, opencl_context, sample_mag, bit_pick, init_bit_pick
 
 
 epsilon = opencl_context.epsilon
@@ -64,26 +64,27 @@ def local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, re
 def compute_energy(sample, G_func, nodes, n_qubits):
     energy = 0
     for u in range(n_qubits):
+        u_bit = sample[u]
         for v in range(u + 1, n_qubits):
             val = G_func(nodes[u], nodes[v])
-            energy += val if sample[u] == sample[v] else -val
+            energy += val if u_bit == sample[v] else -val
 
     return -energy
 
 
 @njit
 def compute_cut(sample, G_func, nodes, n_qubits):
+    l, r = get_cut_base(sample, n_qubits)
     cut = 0
-    for u in range(n_qubits):
-        for v in range(u + 1, n_qubits):
-            if sample[u] != sample[v]:
-                cut += G_func(nodes[u], nodes[v])
+    for u in l:
+        for v in r:
+            cut += G_func(nodes[u], nodes[v])
 
     return cut
 
 
 @njit(parallel=True)
-def sample_measurement(G_func, nodes, max_edge, shots, thresholds, degrees_sum, weights, n, repulsion_base, is_spin_glass):
+def sample_measurement(G_func, nodes, max_edge, shots, thresholds, weights, n, repulsion_base, is_spin_glass):
     shots = max(1, shots >> 1)
     tot_init_weight = weights.sum()
 
@@ -96,14 +97,24 @@ def sample_measurement(G_func, nodes, max_edge, shots, thresholds, degrees_sum,
     improved = True
     while improved:
         improved = False
-        for s in prange(shots):
-            # First dimension: Hamming weight
-            m = sample_mag(thresholds)
+        if is_spin_glass:
+            for s in prange(shots):
+                # First dimension: Hamming weight
+                m = sample_mag(thresholds)
 
-            # Second dimension: permutation within Hamming weight
-            sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, repulsion_base, n, m)
-            solutions[s] = sample
-            energies[s] = compute_energy(sample, G_func, nodes, n) if is_spin_glass else compute_cut(sample, G_func, nodes, n)
+                # Second dimension: permutation within Hamming weight
+                sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, repulsion_base, n, m)
+                solutions[s] = sample
+                energies[s] = compute_energy(sample, G_func, nodes, n)
+        else:
+            for s in prange(shots):
+                # First dimension: Hamming weight
+                m = sample_mag(thresholds)
+
+                # Second dimension: permutation within Hamming weight
+                sample = local_repulsion_choice(G_func, nodes, max_edge, weights, tot_init_weight, repulsion_base, n, m)
+                solutions[s] = sample
+                energies[s] = compute_cut(sample, G_func, nodes, n)
 
         best_index = np.argmax(energies)
         energy = energies[best_index]
@@ -119,7 +130,7 @@ def sample_measurement(G_func, nodes, max_edge, shots, thresholds, degrees_sum,
 
 
 @njit(parallel=True)
-def init_J_and_z(G_func, nodes):
+def init_J_and_z(G_func, nodes, G_min):
     n_qubits = len(nodes)
     degrees = np.empty(n_qubits, dtype=np.uint32)
     J_eff = np.empty(n_qubits, dtype=np.float64)
@@ -129,31 +140,44 @@ def init_J_and_z(G_func, nodes):
         J = 0.0
         for m in range(n_qubits):
             val = G_func(nodes[n], nodes[m])
+            if val > G_max:
+                G_max = val
+            val -= G_min
             if val != 0.0:
                 degree += 1
                 J += val
                 val = abs(val)
-            if val > G_max:
-                G_max = val
         J = -J / degree if degree > 0 else 0
         degrees[n] = degree
         J_eff[n] = J
 
     return J_eff, degrees, G_max
 
+@njit
+def find_G_min(G_func, nodes, n_nodes):
+    G_min = float("inf")
+    for i in range(n_nodes):
+        u = nodes[i]
+        for j in range(i + 1, n_nodes):
+            v = nodes[j]
+            val = G_func(u, v)
+            if val < G_min:
+                G_min = val
+
+    return G_min
+
 
 @njit
-def cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base):
-    J_eff, degrees, max_edge = init_J_and_z(G_func, nodes)
+def cpu_footer(shots, quality, n_qubits, G_min, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base):
+    J_eff, degrees, max_edge = init_J_and_z(G_func, nodes, G_min)
     hamming_prob = maxcut_hamming_cdf(n_qubits, J_eff, degrees, quality, anneal_t, anneal_h)
-    degrees_sum = degrees.sum()
 
     degrees = None
-    J_eff = 1.0 / (1.0 + epsilon - J_eff)
+    J_eff = repulsion_base ** J_eff
 
-    best_solution, best_value = sample_measurement(G_func, nodes, max_edge, shots, hamming_prob, degrees_sum, J_eff, n_qubits, repulsion_base, is_spin_glass)
+    best_solution, best_value = sample_measurement(G_func, nodes, max_edge, shots, hamming_prob, J_eff, n_qubits, repulsion_base, is_spin_glass)
 
-    bit_string, l, r = get_cut(best_solution, nodes)
+    bit_string, l, r = get_cut(best_solution, nodes, n_qubits)
 
     return bit_string, best_value, (l, r)
 
@@ -186,7 +210,7 @@ def maxcut_tfim_streaming(
             return "01", weight, ([nodes[0]], [nodes[1]])
 
     if quality is None:
-        quality = 6
+        quality = 5
 
     if shots is None:
         # Number of measurement shots
@@ -201,7 +225,9 @@ def maxcut_tfim_streaming(
     if repulsion_base is None:
         repulsion_base = 8.0
 
-    bit_string, best_value, partition = cpu_footer(shots, quality, n_qubits, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base)
+    G_min = find_G_min(G_func, nodes, n_qubits)
+
+    bit_string, best_value, partition = cpu_footer(shots, quality, n_qubits, G_min, G_func, nodes, is_spin_glass, anneal_t, anneal_h, repulsion_base)
 
     if best_value < 0.0:
         # Best cut is trivial partition, all/empty
