@@ -7,7 +7,7 @@ from scipy.sparse import lil_matrix
 
 
 class OpenCLContext:
-    def __init__(self, p, a, w, d, e, r, c, q, b, s, x, y):
+    def __init__(self, p, a, w, d, e, r, c, q, b, s, x, y, i, j, k, l):
         self.MAX_GPU_PROC_ELEM = p
         self.IS_OPENCL_AVAILABLE = a
         self.work_group_size = w
@@ -20,10 +20,10 @@ class OpenCLContext:
         self.bootstrap_sparse_kernel = s
         self.bootstrap_segmented_kernel = x
         self.bootstrap_sparse_segmented_kernel = y
-        self.G_m_buf = None
-        self.G_data_buf = None
-        self.G_rows_buf = None
-        self.G_cols_buf = None
+        self.calculate_cut_kernel = i
+        self.calculate_cut_sparse_kernel = j
+        self.calculate_cut_segmented_kernel = k
+        self.calculate_cut_sparse_segmented_kernel = l
 
 IS_OPENCL_AVAILABLE = True
 ctx = None
@@ -37,6 +37,10 @@ bootstrap_kernel = None
 bootstrap_sparse_kernel = None
 bootstrap_segmented_kernel = None
 bootstrap_sparse_segmented_kernel = None
+calculate_cut_kernel = None
+calculate_cut_sparse_kernel = None
+calculate_cut_segmented_kernel = None
+calculate_cut_sparse_segmented_kernel = None
 
 dtype_bits = int(os.getenv('PYQRACKISING_FPPOW', '5'))
 kernel_src = ''
@@ -83,14 +87,16 @@ try:
     compute_units = int(os.getenv('PYQRACKISING_MAX_GPU_PROC_ELEM', str(ctx.devices[0].get_info(cl.device_info.MAX_COMPUTE_UNITS))))
 
     # Load and build OpenCL kernels
-    kernel_src += f"#define MAX_PROC_ELEM {compute_units}\n"
-    kernel_src += f"#define TOP_N {os.getenv('PYQRACKISING_GPU_TOP_N', '32')}\n"
     kernel_src += open(os.path.dirname(os.path.abspath(__file__)) + "/kernels.cl").read()
     program = cl.Program(ctx, kernel_src).build()
     bootstrap_kernel = program.bootstrap
     bootstrap_sparse_kernel = program.bootstrap_sparse
     bootstrap_segmented_kernel = program.bootstrap_segmented
     bootstrap_sparse_segmented_kernel = program.bootstrap_sparse_segmented
+    calculate_cut_kernel = program.calculate_cut
+    calculate_cut_sparse_kernel = program.calculate_cut_sparse
+    calculate_cut_segmented_kernel = program.calculate_cut_segmented
+    calculate_cut_sparse_segmented_kernel = program.calculate_cut_sparse_segmented
 
     work_group_size = bootstrap_kernel.get_work_group_info(
         cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE,
@@ -102,14 +108,119 @@ except ImportError:
     IS_OPENCL_AVAILABLE = False
     print("PyOpenCL not installed. (If you have any OpenCL accelerator devices with available ICDs, you might want to optionally install pyopencl.)")
 
-opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, epsilon, max_alloc, ctx, queue, bootstrap_kernel, bootstrap_sparse_kernel, bootstrap_segmented_kernel, bootstrap_sparse_segmented_kernel)
+opencl_context = OpenCLContext(compute_units, IS_OPENCL_AVAILABLE, work_group_size, dtype, epsilon, max_alloc, ctx, queue, bootstrap_kernel, bootstrap_sparse_kernel, bootstrap_segmented_kernel, bootstrap_sparse_segmented_kernel, calculate_cut_kernel, calculate_cut_sparse_kernel, calculate_cut_segmented_kernel, calculate_cut_sparse_segmented_kernel)
+
+
+def setup_opencl(l, g, args_np):
+    ctx = opencl_context.ctx
+    queue = opencl_context.queue
+    dtype = opencl_context.dtype
+    wgs = opencl_context.work_group_size
+
+    # Buffers
+    mf = cl.mem_flags
+    args_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args_np)
+
+    # Group sizes
+    local_size = min(wgs, l)
+    max_global_size = ((opencl_context.MAX_GPU_PROC_ELEM + local_size - 1) // local_size) * local_size  # corresponds to MAX_PROC_ELEM macro in OpenCL kernel program
+    global_size = min(((g + local_size - 1) // local_size) * local_size, max_global_size)
+
+    # Local memory allocation (1 float per work item)
+    local_energy_buf = cl.LocalMemory(np.dtype(dtype).itemsize * local_size)
+    local_index_buf = cl.LocalMemory(np.dtype(np.int32).itemsize * local_size)
+
+    # Allocate max_energy and max_index result buffers per workgroup
+    max_energy_host = np.empty(global_size // local_size, dtype=dtype)
+    max_index_host = np.empty(global_size // local_size, dtype=np.int32)
+
+    max_energy_buf = cl.Buffer(ctx, mf.WRITE_ONLY, max_energy_host.nbytes)
+    max_index_buf = cl.Buffer(ctx, mf.WRITE_ONLY, max_index_host.nbytes)
+
+    return local_size, global_size, args_buf, local_energy_buf, local_index_buf, max_energy_host, max_index_host, max_energy_buf, max_index_buf
+
+
+def make_G_m_buf(G_m, is_segmented, segment_size):
+    mf = cl.mem_flags
+    ctx = opencl_context.ctx
+    if is_segmented:
+        o_shape = G_m.shape[0] * G_m.shape[1]
+        n_shape = segment_size << 2
+        _G_m = np.reshape(G_m, (o_shape,))
+        if n_shape != o_shape:
+            np.resize(_G_m, (n_shape,))
+        G_m_segments = np.split(_G_m, 4)
+        G_m_buf = [
+            cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=seg)
+            for seg in G_m_segments
+        ]
+    else:
+        G_m_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m)
+
+    return G_m_buf
+
+
+def make_G_m_csr_buf(G_m, is_segmented, segment_size):
+    mf = cl.mem_flags
+    ctx = opencl_context.ctx
+    G_rows_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m.indptr)
+    G_cols_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m.indices)
+    if is_segmented:
+        o_shape = G_m.data.shape[0]
+        n_shape = segment_size << 2
+        _G_data = np.reshape(G_m.data, (o_shape,))
+        if n_shape != o_shape:
+            np.resize(_G_data, (n_shape,))
+        G_data_segments = np.split(_G_data, 4)
+        G_data_buf = [
+            cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=seg)
+            for seg in G_data_segments
+        ]
+        _G_data = None
+        _G_data_segments = None
+    else:
+        G_data_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m.data)
+
+    return G_data_buf, G_rows_buf, G_cols_buf
+
+
+def make_theta_buf(theta, is_segmented, shots, n):
+    mf = cl.mem_flags
+    ctx = opencl_context.ctx
+    if is_segmented:
+        n_shape = (((shots + 3) >> 2) << 2) * ((n + 31) >> 5)
+        theta = np.reshape(theta, (n_shape,))
+        theta_segments = np.split(theta, 4)
+        theta_buf = [
+            cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=seg)
+            for seg in theta_segments
+        ]
+    else:
+        theta_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=theta)
+
+    return theta_buf
+
+
+@njit(parallel=True)
+def convert_bool_to_uint(samples):
+    shots = samples.shape[0]
+    n = samples.shape[1]
+    n32 = (n + 31) >> 5
+    theta = np.zeros(shots * n32, dtype=np.uint32)
+    for i in prange(shots):
+        i_offset = i * n32
+        for j in range(n):
+            if samples[i, j]:
+                theta[i_offset + (j >> 5)] |= 1 << (j & 31)
+
+    return theta
 
 
 @njit
-def get_cut(solution, nodes):
+def get_cut(solution, nodes, n):
     bit_string = ""
     l, r = [], []
-    for i in range(len(solution)):
+    for i in range(n):
         if solution[i]:
             bit_string += "1"
             r.append(nodes[i])
@@ -120,6 +231,18 @@ def get_cut(solution, nodes):
     return bit_string, l, r
 
 
+@njit
+def get_cut_base(solution, n):
+    l, r = [], []
+    for i in range(n):
+        if solution[i]:
+            r.append(i)
+        else:
+            l.append(i)
+
+    return l, r
+
+
 @njit
 def binary_search(l, t):
     left = 0
@@ -239,23 +362,18 @@ def maxcut_hamming_cdf(n_qubits, J_func, degrees, quality, tot_t, h_mult):
         for i in range(n_bias):
             hamming_prob[i] += bias[i] - last_bias[i]
 
-    cum_prob = fix_cdf(hamming_prob)
-
-    return cum_prob
-
-@njit
-def fix_cdf(hamming_prob):
     hamming_prob /= hamming_prob.sum() - (hamming_prob[0] + hamming_prob[-1])
     tot_prob = 0.0
-    n_bias = len(hamming_prob) - 2
+    n_bias -= 2
     cum_prob = np.empty(n_bias, dtype=np.float64)
     for i in range(n_bias):
         tot_prob += hamming_prob[i + 1]
         cum_prob[i] = tot_prob
-    cum_prob[-1] = 2.0
+    cum_prob[-1] = 1.0
 
     return cum_prob
 
+
 @njit
 def sample_mag(cum_prob):
     p = np.random.random()
@@ -276,13 +394,13 @@ def sample_mag(cum_prob):
     return m + 1
 
 @njit
-def init_bit_pick(weights, tot_init_weight, n):
-    r = np.random.rand()
+def init_bit_pick(weights, p, n):
+    p *= np.random.rand()
     cum = 0.0
     node = 0
     for i in range(n):
         cum += weights[i]
-        if (tot_init_weight * r) < cum:
+        if p < cum:
             node = i
             break
 
@@ -291,27 +409,22 @@ def init_bit_pick(weights, tot_init_weight, n):
 @njit
 def bit_pick(weights, used, n):
     # Count available
-    total_w = 0.0
+    p = 0.0
     for i in range(n):
         if used[i]:
             continue
-        total_w += weights[i]
+        p += weights[i]
 
     # Normalize & sample
-    r = np.random.rand()
+    p *= np.random.rand()
     cum = 0.0
-    node = -1
+    node = 0
     for i in range(n):
         if used[i]:
             continue
         cum += weights[i]
-        if (total_w * r) < cum:
+        if p < cum:
             node = i
             break
 
-    if node == -1:
-        node = 0
-        while used[node]:
-            node += 1
-
     return node
