@@ -1,7 +1,8 @@
 from .maxcut_tfim import maxcut_tfim
-from .maxcut_tfim_util import opencl_context
+from .maxcut_tfim_util import get_cut_base, make_G_m_buf, opencl_context, setup_opencl
 from .spin_glass_solver_util import get_cut_from_bit_array, int_to_bitstring
 import itertools
+import math
 import networkx as nx
 import numpy as np
 from numba import njit, prange
@@ -17,45 +18,43 @@ except ImportError:
 
 
 @njit
-def evaluate_cut_edges(theta_bits, G_m):
-    n_qubits = len(G_m)
+def evaluate_cut_edges(theta_bits, G_m, n):
+    l, r = get_cut_base(theta_bits, n)
     cut = 0
-    for u in range(n_qubits):
-        for v in range(u + 1, n_qubits):
-            if theta_bits[u] != theta_bits[v]:
-                cut += G_m[u, v]
+    for u in l:
+        for v in r:
+            cut += G_m[u, v]
 
     return cut
 
 
 @njit
-def compute_energy(theta_bits, G_m):
-    n_qubits = len(G_m)
+def compute_energy(theta_bits, G_m, n):
     energy = 0
-    for u in range(n_qubits):
-        for v in range(u + 1, n_qubits):
-            energy += G_m[u, v] if theta_bits[u] == theta_bits[v] else -G_m[u, v]
+    for u in range(n):
+        u_bit = theta_bits[u]
+        for v in range(u + 1, n):
+            energy += G_m[u, v] if u_bit == theta_bits[v] else -G_m[u, v]
 
     return energy
 
 
 @njit
-def bootstrap_worker(theta, G_m, indices, is_spin_glass):
+def bootstrap_worker(theta, G_m, indices, is_spin_glass, n):
     local_theta = theta.copy()
     for i in indices:
         local_theta[i] = not local_theta[i]
-    energy = compute_energy(local_theta, G_m) if is_spin_glass else -evaluate_cut_edges(local_theta, G_m)
+    energy = compute_energy(local_theta, G_m, n) if is_spin_glass else -evaluate_cut_edges(local_theta, G_m, n)
 
     return energy
 
 
 @njit(parallel=True)
-def bootstrap(best_theta, G_m, indices_array, k, min_energy, dtype, is_spin_glass):
-    n = len(indices_array) // k
+def bootstrap(best_theta, G_m, indices_array, k, min_energy, dtype, is_spin_glass, n):
     energies = np.empty(n, dtype=dtype)
     for i in prange(n):
         j = i * k
-        energies[i] = bootstrap_worker(best_theta, G_m, indices_array[j : j + k], is_spin_glass)
+        energies[i] = bootstrap_worker(best_theta, G_m, indices_array[j : j + k], is_spin_glass, n)
 
     energy = energies.min()
     if energy < min_energy:
@@ -68,46 +67,23 @@ def bootstrap(best_theta, G_m, indices_array, k, min_energy, dtype, is_spin_glas
     return min_energy
 
 
-def run_bootstrap_opencl(best_theta, G_m_buf, indices_array_np, k, min_energy, is_segmented, segment_size, is_spin_glass):
+def run_bootstrap_opencl(best_theta, G_m_buf, indices_array_np, k, min_energy, is_segmented, local_size, global_size, args_buf, local_energy_buf, local_index_buf, max_energy_host, max_index_host, max_energy_buf, max_index_buf):
     ctx = opencl_context.ctx
     queue = opencl_context.queue
     bootstrap_kernel = opencl_context.bootstrap_segmented_kernel if is_segmented else opencl_context.bootstrap_kernel
     dtype = opencl_context.dtype
     epsilon = opencl_context.epsilon
-    wgs = opencl_context.work_group_size
-
-    n = best_theta.shape[0]
-    combo_count = len(indices_array_np) // k
 
     best_theta_np = np.array([(1 if b else 0) for b in best_theta], dtype=np.int8)
 
-    # Args: [n, k, combo_count, segment_size]
-    args_np = np.array([n, k, combo_count, is_spin_glass, segment_size], dtype=np.int32)
-
     # Buffers
     mf = cl.mem_flags
     best_theta_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=best_theta_np)
     indices_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=indices_array_np)
-    args_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=args_np)
-
-    # Local memory allocation (1 float per work item)
-    local_size = min(wgs, n)
-    max_global_size = ((opencl_context.MAX_GPU_PROC_ELEM + local_size - 1) // local_size) * local_size  # corresponds to MAX_PROC_ELEM macro in OpenCL kernel program
-    global_size = min(((combo_count + local_size - 1) // local_size) * local_size, max_global_size)
-    local_energy_buf = cl.LocalMemory(np.dtype(dtype).itemsize * local_size)
-    local_index_buf = cl.LocalMemory(np.dtype(np.int32).itemsize * local_size)
-
-    # Allocate min_energy and min_index result buffers per workgroup
-    min_energy_host = np.empty(global_size, dtype=dtype)
-    min_index_host = np.empty(global_size, dtype=np.int32)
-
-    min_energy_buf = cl.Buffer(ctx, mf.WRITE_ONLY, min_energy_host.nbytes)
-    min_index_buf = cl.Buffer(ctx, mf.WRITE_ONLY, min_index_host.nbytes)
 
     # Set kernel args
     if is_segmented:
         bootstrap_kernel.set_args(
-            np.random.randint(1<<32, dtype='uint32'),
             G_m_buf[0],
             G_m_buf[1],
             G_m_buf[2],
@@ -115,20 +91,19 @@ def run_bootstrap_opencl(best_theta, G_m_buf, indices_array_np, k, min_energy, i
             best_theta_buf,
             indices_buf,
             args_buf,
-            min_energy_buf,
-            min_index_buf,
+            max_energy_buf,
+            max_index_buf,
             local_energy_buf,
             local_index_buf
         )
     else:
         bootstrap_kernel.set_args(
-            np.random.randint(1<<32, dtype='uint32'),
             G_m_buf,
             best_theta_buf,
             indices_buf,
             args_buf,
-            min_energy_buf,
-            min_index_buf,
+            max_energy_buf,
+            max_index_buf,
             local_energy_buf,
             local_index_buf
         )
@@ -136,19 +111,16 @@ def run_bootstrap_opencl(best_theta, G_m_buf, indices_array_np, k, min_energy, i
     cl.enqueue_nd_range_kernel(queue, bootstrap_kernel, (global_size,), (local_size,))
 
     # Read results
-    cl.enqueue_copy(queue, min_energy_host, min_energy_buf)
-    cl.enqueue_copy(queue, min_index_host, min_index_buf)
+    cl.enqueue_copy(queue, max_energy_host, max_energy_buf)
+    cl.enqueue_copy(queue, max_index_host, max_index_buf)
     queue.finish()
 
     # Find global minimum
-    energy = min_energy_host.min()
-    if min_energy < energy:
-        return min_energy
+    best_i = np.argmax(max_energy_host)
+    best_energy = -max_energy_host[best_i]
 
-    atol = dtype(epsilon)
-    rtol = dtype(0)
-    choices = np.where(np.isclose(min_energy_host, energy, atol=atol, rtol=rtol))[0]
-    best_i = np.random.choice(choices) if len(choices) else np.argmin(min_energy_host)
+    if min_energy <= best_energy:
+        return min_energy
 
     flip_index_start = best_i * k
     indices_to_flip = indices_array_np[flip_index_start : flip_index_start + k]
@@ -156,16 +128,16 @@ def run_bootstrap_opencl(best_theta, G_m_buf, indices_array_np, k, min_energy, i
     for i in indices_to_flip:
         best_theta[i] = not best_theta[i]
 
-    return min_energy_host[best_i]
+    return best_energy
 
 
-def log_intermediate(theta, G_m, nodes, min_energy, is_spin_glass):
+def log_intermediate(theta, G_m, nodes, min_energy, is_spin_glass, n):
     bitstring, l, r = get_cut_from_bit_array(theta, nodes)
     if is_spin_glass:
-        cut_value = evaluate_cut_edges(theta, G_m)
+        cut_value = evaluate_cut_edges(theta, G_m, n)
     else:
         cut_value = -min_energy
-        min_energy = compute_energy(theta, G_m)
+        min_energy = compute_energy(theta, G_m, n)
     print(bitstring, float(cut_value), (l, r), float(min_energy))
 
 
@@ -179,8 +151,8 @@ def spin_glass_solver(
     anneal_t=None,
     anneal_h=None,
     repulsion_base=None,
-    min_order=1,
-    max_order=None,
+    min_order=2,
+    max_order=2,
     is_log=False,
     reheat_tries=0
 ):
@@ -197,12 +169,6 @@ def spin_glass_solver(
         nodes = list(range(n_qubits))
         G_m = G
 
-    segment_size = G_m.shape[0] ** 2
-    is_segmented = (G_m.nbytes << 1) > opencl_context.max_alloc
-    if is_segmented and is_alt_gpu_sampling:
-        print("[WARN] Using segmented solver, so disabling is_alt_gpu_sampling.")
-        is_alt_gpu_sampling = False
-
     if n_qubits < 3:
         if n_qubits == 0:
             return "", 0, ([], []), 0
@@ -225,39 +191,24 @@ def spin_glass_solver(
     elif isinstance(best_guess, list):
         bitstring = "".join(["1" if b else "0" for b in best_guess])
     else:
-        bitstring, _, _ = maxcut_tfim(G_m, quality=quality, shots=shots, is_spin_glass=is_spin_glass, anneal_t=anneal_t, anneal_h=anneal_h, repulsion_base=repulsion_base)
+        bitstring, _, _ = maxcut_tfim(G_m, quality=quality, shots=shots, is_spin_glass=is_spin_glass, anneal_t=anneal_t, anneal_h=anneal_h, repulsion_base=repulsion_base, is_maxcut_gpu=is_combo_maxcut_gpu, is_nested=True)
     best_theta = np.array([b == "1" for b in list(bitstring)], dtype=np.bool_)
 
-    if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
-        if not (opencl_context.G_m_buf is None):
-            G_m_buf = opencl_context.G_m_buf
-        else:
-            mf = cl.mem_flags
-            ctx = opencl_context.ctx
-            if is_segmented:
-                o_shape = segment_size
-                segment_size = (segment_size + 3) >> 2
-                n_shape = segment_size << 2
-                _G_m = np.reshape(G_m, (o_shape,))
-                if n_shape != o_shape:
-                    np.resize(_G_m, (n_shape,))
-                _G_m_segments = np.split(_G_m, 4)
-                G_m_buf = [
-                    cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=seg)
-                    for seg in _G_m_segments
-                ]
-                _G_m = None
-                _G_m_segments = None
-            else:
-                G_m_buf = cl.Buffer(ctx, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=G_m)
+    segment_size = G_m.shape[0] * G_m.shape[1]
+    is_segmented = (G_m.nbytes << 1) > opencl_context.max_alloc
+
+    is_opencl = is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE
+
+    if is_opencl:
+        G_m_buf = make_G_m_buf(G_m, is_segmented, segment_size)
 
     if max_order is None:
         max_order = n_qubits
 
-    min_energy = compute_energy(best_theta, G_m) if is_spin_glass else -evaluate_cut_edges(best_theta, G_m)
+    min_energy = compute_energy(best_theta, G_m, n_qubits) if is_spin_glass else -evaluate_cut_edges(best_theta, G_m, n_qubits)
 
     if is_log:
-        log_intermediate(best_theta, G_m, nodes, min_energy, is_spin_glass)
+        log_intermediate(best_theta, G_m, nodes, min_energy, is_spin_glass, n_qubits)
 
     combos_list = []
     reheat_theta = best_theta.copy()
@@ -280,10 +231,12 @@ def spin_glass_solver(
                 else:
                     combos = combos_list[k - 1]
 
-                if is_combo_maxcut_gpu and IS_OPENCL_AVAILABLE:
-                    energy = run_bootstrap_opencl(reheat_theta, G_m_buf, combos, k, reheat_min_energy, is_segmented, segment_size, is_spin_glass)
+                if is_opencl:
+                    combo_count = len(combos) // k
+                    opencl_args = setup_opencl(n_qubits, combo_count, np.array([n_qubits, k, combo_count, is_spin_glass, segment_size], dtype=np.int32))
+                    energy = run_bootstrap_opencl(reheat_theta, G_m_buf, combos, k, reheat_min_energy, is_segmented, *opencl_args)
                 else:
-                    energy = bootstrap(reheat_theta, G_m, combos, k, reheat_min_energy, dtype, is_spin_glass)
+                    energy = bootstrap(reheat_theta, G_m, combos, k, reheat_min_energy, dtype, is_spin_glass, n_qubits)
 
                 if energy < reheat_min_energy:
                     reheat_min_energy = energy
@@ -291,7 +244,7 @@ def spin_glass_solver(
                     if correction_quality < (k + 1):
                         correction_quality = k + 1
                     if is_log:
-                        log_intermediate(reheat_theta, G_m, nodes, reheat_min_energy, is_spin_glass)
+                        log_intermediate(reheat_theta, G_m, nodes, reheat_min_energy, is_spin_glass, n_qubits)
 
                     break
 
@@ -308,15 +261,15 @@ def spin_glass_solver(
             bits_to_flip = random.sample(list(range(n_qubits)), num_to_flip)
             for bit in bits_to_flip:
                 reheat_theta[bit] = not reheat_theta[bit]
-            reheat_min_energy = compute_energy(reheat_theta, G_m) if is_spin_glass else -evaluate_cut_edges(reheat_theta, G_m)
+            reheat_min_energy = compute_energy(reheat_theta, G_m, n_qubits) if is_spin_glass else -evaluate_cut_edges(reheat_theta, G_m, n_qubits)
 
     opencl_context.G_m_buf = None
 
     bitstring, l, r = get_cut_from_bit_array(best_theta, nodes)
     if is_spin_glass:
-        cut_value = evaluate_cut_edges(best_theta, G_m)
+        cut_value = evaluate_cut_edges(best_theta, G_m, n_qubits)
     else:
         cut_value = -min_energy
-        min_energy = compute_energy(best_theta, G_m)
+        min_energy = compute_energy(best_theta, G_m, n_qubits)
 
     return bitstring, float(cut_value), (l, r), float(min_energy)
