@@ -1,5 +1,5 @@
 from .maxcut_tfim_streaming import maxcut_tfim_streaming
-from .maxcut_tfim_util import opencl_context
+from .maxcut_tfim_util import get_cut_base, opencl_context
 from .spin_glass_solver_util import get_cut_from_bit_array, int_to_bitstring
 import itertools
 import networkx as nx
@@ -13,46 +13,44 @@ epsilon = opencl_context.epsilon
 
 
 @njit
-def evaluate_cut_edges(theta_bits, G_func, nodes):
-    n_qubits = len(nodes)
+def evaluate_cut_edges(theta_bits, G_func, nodes, n):
+    l, r = get_cut_base(theta_bits, n)
     cut = 0
-    for u in range(n_qubits):
-        for v in range(u + 1, n_qubits):
-            if theta_bits[u] != theta_bits[v]:
-                cut += G_func(nodes[u], nodes[v])
+    for u in l:
+        for v in r:
+            cut += G_func(nodes[u], nodes[v])
 
     return cut
 
 
 @njit
-def compute_energy(theta_bits, G_func, nodes):
-    n_qubits = len(nodes)
+def compute_energy(theta_bits, G_func, nodes, n):
     energy = 0
-    for u in range(n_qubits):
-        for v in range(u + 1, n_qubits):
+    for u in range(n):
+        u_bit = theta_bits[u]
+        for v in range(u + 1, n):
             val = G_func(nodes[u], nodes[v])
-            energy += (val if theta_bits[u] == theta_bits[v] else -val)
+            energy += (val if u_bit == theta_bits[v] else -val)
 
     return energy
 
 
 @njit
-def bootstrap_worker(theta, G_func, nodes, indices, is_spin_glass):
+def bootstrap_worker(theta, G_func, nodes, indices, is_spin_glass, n):
     local_theta = theta.copy()
     for i in indices:
         local_theta[i] = not local_theta[i]
-    energy = compute_energy(local_theta, G_func, nodes) if is_spin_glass else -evaluate_cut_edges(local_theta, G_func, nodes)
+    energy = compute_energy(local_theta, G_func, nodes, n) if is_spin_glass else -evaluate_cut_edges(local_theta, G_func, nodes, n)
 
     return energy
 
 
 @njit(parallel=True)
-def bootstrap(best_theta, G_func, nodes, indices_array, k, min_energy, dtype, is_spin_glass):
-    n = len(indices_array) // k
+def bootstrap(best_theta, G_func, nodes, indices_array, k, min_energy, dtype, is_spin_glass, n):
     energies = np.empty(n, dtype=dtype)
     for i in prange(n):
         j = i * k
-        energies[i] = bootstrap_worker(best_theta, G_func, nodes, indices_array[j : j + k], is_spin_glass)
+        energies[i] = bootstrap_worker(best_theta, G_func, nodes, indices_array[j : j + k], is_spin_glass, n)
 
     energy = energies.min()
     if energy < min_energy:
@@ -67,13 +65,13 @@ def bootstrap(best_theta, G_func, nodes, indices_array, k, min_energy, dtype, is
     return min_energy
 
 
-def log_intermediate(theta, G_func, nodes, min_energy, is_spin_glass):
+def log_intermediate(theta, G_func, nodes, min_energy, is_spin_glass, n):
     bitstring, l, r = get_cut_from_bit_array(theta, nodes)
     if is_spin_glass:
-        cut_value = evaluate_cut_edges(theta, G_func, nodes)
+        cut_value = evaluate_cut_edges(theta, G_func, nodes, n)
     else:
         cut_value = -min_energy
-        min_energy = compute_energy(theta, G_func, nodes)
+        min_energy = compute_energy(theta, G_func, nodes, n)
     print(bitstring, float(cut_value), (l, r), float(min_energy))
 
 
@@ -87,8 +85,8 @@ def spin_glass_solver_streaming(
     anneal_t=None,
     anneal_h=None,
     repulsion_base=None,
-    min_order=1,
-    max_order=None,
+    min_order=2,
+    max_order=2,
     is_log=False,
     reheat_tries=0
 ):
@@ -123,10 +121,10 @@ def spin_glass_solver_streaming(
     if max_order is None:
         max_order = n_qubits
 
-    min_energy = compute_energy(best_theta, G_func, nodes) if is_spin_glass else -evaluate_cut_edges(best_theta, G_func, nodes)
+    min_energy = compute_energy(best_theta, G_func, nodes, n_qubits) if is_spin_glass else -evaluate_cut_edges(best_theta, G_func, nodes, n_qubits)
 
     if is_log:
-        log_intermediate(best_theta, G_func, nodes, min_energy, is_spin_glass)
+        log_intermediate(best_theta, G_func, nodes, min_energy, is_spin_glass, n_qubits)
 
     combos_list = []
     reheat_theta = best_theta.copy()
@@ -149,7 +147,7 @@ def spin_glass_solver_streaming(
                 else:
                     combos = combos_list[k - 1]
 
-                energy = bootstrap(best_theta, G_func, nodes, combos, k, min_energy, dtype, is_spin_glass)
+                energy = bootstrap(best_theta, G_func, nodes, combos, k, min_energy, dtype, is_spin_glass, n_qubits)
 
                 if energy < reheat_min_energy:
                     reheat_min_energy = energy
@@ -158,7 +156,7 @@ def spin_glass_solver_streaming(
                         correction_quality = k + 1
 
                     if is_log:
-                        log_intermediate(reheat_theta, G_func, nodes, min_energy, is_spin_glass)
+                        log_intermediate(reheat_theta, G_func, nodes, min_energy, is_spin_glass, n_qubits)
 
                     break
 
@@ -175,13 +173,13 @@ def spin_glass_solver_streaming(
             bits_to_flip = random.sample(list(range(n_qubits)), num_to_flip)
             for bit in bits_to_flip:
                 reheat_theta[bit] = not reheat_theta[bit]
-            reheat_min_energy = compute_energy(reheat_theta, G_func, nodes) if is_spin_glass else -evaluate_cut_edges(reheat_theta, G_func, nodes)
+            reheat_min_energy = compute_energy(reheat_theta, G_func, nodes, n_qubits) if is_spin_glass else -evaluate_cut_edges(reheat_theta, G_func, nodes, n_qubits)
 
     bitstring, l, r = get_cut_from_bit_array(best_theta, nodes)
     if is_spin_glass:
-        cut_value = evaluate_cut_edges(best_theta, G_func, nodes)
+        cut_value = evaluate_cut_edges(best_theta, G_func, nodes, n_qubits)
     else:
         cut_value = -min_energy
-        min_energy = compute_energy(best_theta, G_func, nodes)
+        min_energy = compute_energy(best_theta, G_func, nodes, n_qubits)
 
     return bitstring, float(cut_value), (l, r), float(min_energy)
